import os
import mathutils

def linear_to_srgb(l):
	if l<0.0031308:
		return 12.92*l
	else:
		return 1.055*(l**(1/2.4))-0.055

def image_name(i):
	fp = i.filepath
	if fp:
		return os.path.split(fp)[1]
	else:
		return i.name

def external_name(out_file, ext, index):
	path, base = os.path.split(out_file.filename)
	base = os.path.splitext(base)[0]
	if index>0:
		base += "_lod{}".format(index)
	return path, base+ext


class ObjectExporter:
	def __init__(self):
		self.material_tex = False
		self.srgb_colors = True
		self.textures = "REF"
		self.separate_mesh = False
		self.shared_mesh = True
		self.separate_tech = False
		self.external_tech = True
		self.shared_tech = True
		self.export_lods = True

	def export(self, context, out_file, objs=None, progress=None):
		if objs is None:
			obj = context.active_object
		else:
			obj = objs[0]

		lods = [obj]
		for c in obj.children:
			if c.lod_for_parent:
				if c.lod_index>=len(lods):
					lods += [None]*(c.lod_index+1-len(lods))
				lods[c.lod_index] = c

		from .outfile import open_output
		out_file = open_output(out_file)

		p1 = max(((v.co, v.co.length) for v in obj.data.vertices), key=lambda x:x[1])[0]
		p2 = max(((v.co, (v.co-p1).length) for v in obj.data.vertices), key=lambda x:x[1])[0]
		center = (p1+p2)/2
		radius = (p1-p2).length/2
		for v in obj.data.vertices:
			d = v.co-center
			if d.length>radius:
				center += d*(1-radius/d.length)/2
				radius = (radius+d.length)/2

		out_file.write("bounding_sphere_hint", center[0], center[1], center[2], radius)

		prev_mesh = None
		prev_tech = (None, None)
		for i, l in enumerate(lods):
			if i>0:
				out_file.begin("level_of_detail", i)
				objs = [l]

			if i==0 or l.data.name!=prev_mesh:
				mesh = self.export_object_mesh(context, out_file, l, objs, progress)
				prev_mesh = l.data.name

			same_tech = True
			mat = None
			if l.material_slots and l.material_slots[0].material:
				mat = l.material_slots[0].material.name
				if mat!=prev_tech[1]:
					same_tech = False
			if self.external_tech and l.technique!=prev_tech[0]:
				same_tech = False
			if i==0 or not same_tech:
				self.export_object_technique(l, mesh, out_file, i)
				prev_tech = (l.technique, mat)

			if i>0:
				out_file.end()

	def export_object_mesh(self, context, out_file, lod, objs, progress):
		from .export_mesh import MeshExporter
		mesh_export = MeshExporter()
		for k, v in self.__dict__.items():
			setattr(mesh_export, k, v)

		lod_index = 0
		if lod.lod_for_parent:
			lod_index = lod.lod_index

		if self.separate_mesh:
			from .outfile import open_output
			path, name = external_name(out_file, ".mesh", lod_index)
			if self.shared_mesh:
				name = lod.data.name+".mesh"
			mesh_out = open_output(os.path.join(path, name))
			mesh = mesh_export.export(context, mesh_out, objs, progress)
			out_file.write("mesh", '"{}"'.format(name))
		else:
			out_file.begin("mesh")
			mesh = mesh_export.export(context, out_file, objs, progress)
			out_file.end()

		return mesh

	def export_object_technique(self, obj, mesh, out_file, lod_index):
		if self.srgb_colors:
			self.colormap = linear_to_srgb
		else:
			self.colormap = lambda x: x

		material = None
		if obj.material_slots:
			material = obj.material_slots[0].material

		from .outfile import open_output
		path, name = external_name(out_file, ".tech", lod_index)

		if self.external_tech and obj.technique:
			if obj.inherit_tech and material and (obj.override_material or material.texture_slots):
				out_file.begin("technique")
				out_file.begin("inherit", '"{}"'.format(obj.technique))
				for slot in material.texture_slots:
					if slot and slot.texture.type=="IMAGE":
						name = image_name(slot.texture.image)
						if slot.use_map_color_diffuse:
							out_file.write("texture", '"diffuse_map"', '"{}"'.format(name))
						elif slot.use_map_normal:
							out_file.write("texture", '"normal_map"', '"{}"'.format(name))
				if obj.override_material:
					mat_name = material.name+".mat"
					mat_out = open_output(os.path.join(path, mat_name))
					self.export_material(material, mat_out)
					out_file.write("material", '"surface"', '"{}"'.format(mat_name))
				out_file.end()
				out_file.end()
			else:
				out_file.write("technique", '"{}"'.format(obj.technique))
		elif self.separate_tech:
			if self.shared_tech and material:
				name = material.name+".tech"
			tech_out = open_output(os.path.join(path, name))
			self.export_technique_definition(material, mesh, tech_out)
			out_file.write("technique", '"{}"'.format(name))
		else:
			out_file.begin("technique")
			self.export_technique_definition(material, mesh, out_file)
			out_file.end()

	def export_technique_definition(self, material, mesh, out_file):
		out_file.begin("pass", '""')
		if material:
			cm = self.colormap

			if self.material_tex:
				out_file.begin("material")
				out_file.write("diffuse", 1.0, 1.0, 1.0, 1.0)
				out_file.end()
				index = 0
				for u in mesh.uv_layers:
					if u.name=="material_tex":
						index = u.unit
				out_file.begin("texunit", index)
				out_file.begin("texture2d")
				out_file.write("min_filter", "NEAREST")
				out_file.write("mag_filter", "NEAREST")
				out_file.write("storage", "RGB", len(mesh.materials), 1)
				texdata = '"'
				for m in mesh.materials:
					color = [int(cm(c)*255) for c in m.diffuse_color*mat.diffuse_intensity]
					texdata += "\\x%02X\\x%02X\\x%02X"%tuple(color)
				texdata += '"'
				out_file.write("raw_data", texdata)
				out_file.end()
				out_file.end()
			else:
				out_file.begin("material")
				self.export_material(material, out_file)
				out_file.end()

			if self.textures!="NONE":
				for slot in material.texture_slots:
					if not slot:
						continue

					tex = slot.texture
					if tex.type!="IMAGE":
						continue

					if slot.uv_layer:
						for u in mesh.uv_layers:
							if u.name==slot.uv_layer:
								index = u.unit
					else:
						index = mesh.uv_layers[0].unit

					out_file.begin("texunit", index)
					if self.textures=="INLINE":
						out_file.begin("texture2d")
						out_file.write("min_filter", "LINEAR")
						out_file.write("storage", "RGBA", tex.image.size[0], tex.image.size[1])
						texdata = '"'
						for p in tex.image.pixels:
							texdata += "\\x%02X"%int(p*255)
						texdata += '"'
						out_file.write("raw_data", texdata)
						out_file.end()
					else:
						out_file.write("texture", '"%s"'%image_name(tex.image))
					out_file.end()

		out_file.end()

	def export_material(self, mat, out_file):
		cm = self.colormap
		if any((s and s.use_map_color_diffuse) for s in mat.texture_slots):
			out_file.write("diffuse", 1.0, 1.0, 1.0, 1.0)
			amb = cm(mat.ambient)
			out_file.write("ambient", amb, amb, amb, 1.0)
		else:
			diff = mat.diffuse_color*mat.diffuse_intensity
			out_file.write("diffuse", cm(diff.r), cm(diff.g), cm(diff.b), 1.0)
			amb = diff*mat.ambient
			out_file.write("ambient", cm(amb.r), cm(amb.g), cm(amb.b), 1.0)
		spec = mat.specular_color*mat.specular_intensity
		out_file.write("specular", spec.r, spec.g, spec.b, 1.0)
		out_file.write("shininess", mat.specular_hardness);
